/*
 * Copyright (c) 2014 by ELECTRONIC ASSEMBLY <technik@lcd-module.de>
 * EA OLED-Text (SSD1311) library for arduino.
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1, both as
 * published by the Free Software Foundation.
 */

#include <Arduino.h>
#include <SPI.h>
#include <Wire.h>

#include "uniTFT.h"

unitft uniTFT;

void isr_sendBufferCall()
{
	uniTFT.isr_sendBuffer();
}


//----------------------------------------------------public Functions----------------------------------------------------
//Please use these functions in your sketch

/*----------------------------
Func: initialize_spi
Desc: Initializes SPI Hardware/Software and sets up connection to uniTFTs
Vars: CS-Pin, MOSI-Pin, MISO-Pin, SCK-Pin (MOSI=SCK Hardware else Software), SBUF-Pin (2,3 Pin interrupt for Arduino Uno)
------------------------------*/
void unitft::initialize_spi(byte p_cs, byte p_mosi, byte p_miso, byte p_clk, byte p_sbuf)
{
  initialize(p_sbuf); //General initialisation (SBUF pin interrupt)
  spi_initialize(p_cs, p_mosi, p_miso, p_clk); //init SPI to Mode 3
  spi = true;
  checkReceiveBufferStatus(); //set up small protkoll packet

}

/*----------------------------
Func: initialize_i2c
Desc: Initializes I2C Hardware and sets up connection to uniTFTs
Vars: freqKHz, MOSI-Pin, slaveAddress8bit, SBUF-Pin (2,3 Pin interrupt for Arduino Uno)
------------------------------*/
void unitft::initialize_i2c (int freqKHz, byte slaveAddress8bit, byte p_sbuf)
{
  initialize(p_sbuf); //General initialisation (SBUF pin interrupt)
  i2c_initialize(freqKHz, slaveAddress8bit); //init I2C
  spi = false;
  checkReceiveBufferStatus(); //set up small protkoll packet 
}

/*----------------------------
Func: RegisterCallbackReceive
Desc: set function to be called if data is received from uniTFT
Vars: ---
------------------------------*/
void unitft::RegisterCallbackReceive(ReceiveEvent callbackReceiveData)
{
    reveiveDataCallback = callbackReceiveData;
}

/*----------------------------
Func: sendDataPacket
Desc: sends out a data packet, wrapped in protocoll frames
Vars: *data, length of data
------------------------------*/
boolean unitft::sendDataPacket(const char *dat, unsigned int len)
{
	byte translen;
	while(len)
	{
		translen = (len>252)? 252 : len;
		while(receiveBufferFree<translen)
		{
			checkReceiveBufferStatus();
			if(receiveBufferFree<translen)
				delay(TIMEOUTBUFFERFULL);
		}

		if(sendPacket(dat,translen,DC1) == ACK)
		{
			len -= translen;
			receiveBufferFree -= translen;
			dat +=translen;
		}
		else //Error sending packet
			return false;
	}
	return true;
}

/*----------------------------
Func: isr_sendBuffer
Desc: Receives data from module's send buffer
Vars: ---
------------------------------*/
void unitft::isr_sendBuffer()
{
	detachInterrupt(digitalPinToInterrupt(p_sbuf));
	sei();
	if(sendPacket("S", 1, DC2) == ACK)
	{
		int rec = recPacket();
		if(rec && reveiveDataCallback)
		{
			reveiveDataCallback(rec);
		}
	}
	attachInterrupt(digitalPinToInterrupt(p_sbuf), isr_sendBufferCall, LOW);
}

//----------------------------------------------------private Functions----------------------------------------------------
//normally you don't need those functions in your sketch

/*----------------------------
Func: initialize
Desc: Sets the buffers and attaches the SBUF pin interrupt
Vars: ---
------------------------------*/
void unitft::initialize(byte p_sbuf)
{
	  reveiveDataCallback = NULL;
  if(p_sbuf == 2 || p_sbuf == 3)
  {
	  unitft::p_sbuf = p_sbuf;
	  pinMode(p_sbuf, INPUT);
	  attachInterrupt(digitalPinToInterrupt(p_sbuf), isr_sendBufferCall, LOW);
  }
  
  sendBufferReady = 0;
  receiveBufferFree = 0;
}
  
/*----------------------------
Func: checkReceiveBufferStatus
Desc: Checks how many bytes can be transfered to display
Vars: ---
------------------------------*/
void unitft::checkReceiveBufferStatus()
{
	if(sendPacket("I", 1, DC2) == ACK)
	{
		if(recPacket() == 2)
		{
			sendBufferReady = receivepacket[0];
			receiveBufferFree = receivepacket[1];
		}
	}
}

/*----------------------------
Func: sendPacket
Desc: sends out a protocol packet
Vars: ---
------------------------------*/
byte unitft::sendPacket(const char *dat, byte len, byte type)
{
	if(len>=252)
		return false;
	
	byte *p_send = sendpacket;
	byte sendlen = len;
	byte bcc = type;
	*p_send++ = type;
	*p_send++ = sendlen;
	bcc += sendlen;
	while(len--)
	{
		*p_send++ = *dat;
		bcc += *dat++;
	}
	*p_send = bcc;

	byte ack;
	byte rep = 0;
	do{
		out(sendpacket, sendlen+3);
		delayMicroseconds(6);
		ack = rec();
		rep++;
		if(ack != ACK && ack!= NACK && rep < REPEAT)
			delay(TIMEOUT);
		else if (ack == NACK && rep < REPEAT)
			delay(TIMEOUTBUFFERFULL);
	}while(ack != ACK && rep < REPEAT);
	
	return ack;
}

/*----------------------------
Func: recPacket
Desc: receives data and writes to global receive buffer, return length of valid data
Vars: ---
------------------------------*/
byte unitft::recPacket()
{
	byte *p_dat = receivepacket;
	if(rec(receivepacket, 2) == false) //error
	{
		receivepacket[0] = 0;
		receivepacket[1] = 0;
	}
    byte type = receivepacket[0];
	byte len = receivepacket[1];
	byte counter = len;
	byte bcc = type + len;
	
	rec(receivepacket, counter);
	while(counter--)
	{
		//*p_dat = rec();
		bcc += *p_dat++;
	}
	
	if(bcc == rec())
		return len;
	else
		return 0;
}

/*----------------------------
Func: spi_initialize
Desc: Initializes SPI Hardware/Software
Vars: CS-Pin, MOSI-Pin, MISO-Pin, SCK-Pin (MOSI=SCK Hardware else Software)
------------------------------*/
void unitft::spi_initialize(byte cs, byte mosi, byte miso, byte clk) 
{	
  //Set pin Configuration
  p_cs = cs;
  if(mosi == clk)
  {
    hardware = true;
    p_mosi = MOSI;
	p_miso = MISO;
    p_clk = SCK;

    digitalWrite(SS, HIGH);
    pinMode(SS, OUTPUT);    
  }
  else 
  {
    hardware = false;
    p_mosi = mosi;
	p_miso = miso;	
    p_clk = clk;
  }
  
  // Set CS to deselct slaves
  digitalWrite(p_cs, HIGH);
  pinMode(p_cs, OUTPUT);

  // Set Data pin as output and input
  pinMode(p_mosi, OUTPUT);
  pinMode(p_miso, INPUT);

  // Set SPI-Mode 3: CLK idle high, rising edge, LSB first
  digitalWrite(p_clk, HIGH);
  pinMode(p_clk, OUTPUT);
  if(hardware)
	  SPI.beginTransaction(SPISettings(1000000, MSBFIRST, SPI_MODE3));
}

/*----------------------------
Func: spi_out
Desc: Sends a buffer with length
Vars: buffer, len
------------------------------*/
void unitft::spi_out(byte *dat, byte len)
{ 
  digitalWrite(p_cs, LOW);
  if(hardware) 
  {
     SPI.transfer(dat, len);
  }
  else 
  { 
    while(len--)
	{
		byte i = 8;
		byte trans = *dat++;
		do 
		{
		  if(trans & 0x01)
			 digitalWrite(p_mosi, HIGH);
		  else
			 digitalWrite(p_mosi, LOW);
		  digitalWrite(p_clk, LOW);
		  trans >>= 1;
		digitalWrite(p_clk, HIGH);
		}while(--i);
	}
  }
  digitalWrite(p_cs, HIGH);
}


/*----------------------------
Func: spi_rec
Desc: Sends 0xFF and receives one
Vars: ---
------------------------------*/
byte unitft::spi_rec() 
{
  byte i = 8;
  byte rec;
  digitalWrite(p_cs, LOW);
  if(hardware) 
  {
     rec = SPI.transfer(0xFF);
  }
  else 
  {
	digitalWrite(p_mosi, HIGH);
    do 
	{
	  rec <<= 1;
	  rec |= digitalRead(p_miso);
      digitalWrite(p_clk, HIGH);
    }while(--i);
  }
  digitalWrite(p_cs, HIGH);
  
  return rec;
}

/*----------------------------
Func: spi_rec
Desc: Sends 0xFF and receives multible bytes
Vars: *buffer to store data, length
------------------------------*/
boolean unitft::spi_rec(byte *buf, int len)
{
	while(len--)
	{
		*buf++ = spi_rec();
	}
	return true;
}

/*----------------------------
Func: i2c_initialize
Desc: Initializes I2C Hardware
Vars: Frequency in KHz, 8-Bit slave address
------------------------------*/
void unitft::i2c_initialize(int freqKHz, byte slaveAddress8bit)
{	
  Wire.begin();
  if(freqKHz<10)
	freqKHz=10;
  else if(freqKHz>400)
    freqKHz = 400;
  Wire.setClock((unsigned long)freqKHz*1000);
  
  p_i2cslaveaddress = slaveAddress8bit >> 1;
  
}

/*----------------------------
Func: i2c_out
Desc: Sends a buffer with length
Vars: buffer, len
------------------------------*/
void unitft::i2c_out(byte *dat, byte len)
{ 
  int sendlen;
  while(len)
  {
	sendlen = len;
	if(len>32)
	  sendlen = 32;
	len-=sendlen;
	
	Wire.beginTransmission(p_i2cslaveaddress);
	Wire.write(dat, sendlen); 
	Wire.endTransmission(p_i2cslaveaddress);
	
	dat+=sendlen;
  }

}

/*----------------------------
Func: i2c_rec
Desc: Sends 0xFF and receives one
Vars: ---
------------------------------*/
byte unitft::i2c_rec() 
{
  byte rec = 0x15;
  Wire.requestFrom(p_i2cslaveaddress, 1);
  if(Wire.available())
    rec = Wire.read();
  return rec;
}

/*----------------------------
Func: i2c_rec
Desc: Sends 0xFF and receives multible bytes
Vars: *buffer to store data, length
------------------------------*/
boolean unitft::i2c_rec(byte *buf, int len)
{
  int request;
  while(len)
  {  
    request = len;
	if(request > 32)
	  request = 32;
    Wire.requestFrom(p_i2cslaveaddress, request);
    while(Wire.available())
    {
		len--;
		*buf++ = Wire.read();
    }
  }
  return (len==0);
}
